// Modified by Philip Wilk so a substr w/o a length works

#ifndef _MYSTRING_H
#define _MYSTRING_H

#include <iostream.h>
#include <string.h>            // for strcmp, strlen, etc.
#include <stdlib.h>            // for exit,atof,atoi
#include <assert.h>            // for assert
#include <ctype.h>             // for tolower/toupper

// modeled (loosely) on Stroustrup 2nd edition, on Budd's String class,
// and Adams, Leestma, and Nyhoff Strings (and probably a host of others)
//
// written by Owen Astrachan April 1993 (with reference counts)
//      modified May 2, 1994
//      modified July 6, 1994 [fixed some inconsistencies found in DOS
//                             implementation]
//      modified October, 1994 [to re-use existing memory rather than
//                              than deleting when =, copy constructor used]
//
// provides a "standard" String class supporting operations given below
//
// constructors:
//
//        String(const char * s = 0);
//               copies a C-style '\0'-terminated string
//        String(const String & x);
//               copy constructor
//        String (char ch);
//               make a String from a single character
//
//  destructor
//        ~String()  destructor frees memory associated with String
//
//  operators
//
//        String & operator= (const String & x);
//               assigns one String to another
//        String & operator= (const char * s);
//               assigns a C-style string to a String
//
//        operator const char * s () const
//               cast/convert String to C-style string
//
//        int length() const
//               returns # of characters in String
//
//         char& operator[](int i)
//               returns i-th char of String (signals error if out-of-bounds)
//                       (char is assignable since reference)
//
//         const char& operator[](int i) const
//               returns i-th char of String (signals error if out-of-bounds)
//
//  I/O operators
//       << and >>
//               are "standard" for ostream and istream respectively
//       istream & getline(istream & is, String & s, char sentinel = '\n')
//               mimics iostream.h getline for istreams and Strings
//
//  relational operators
//          ==, !=, <, >, >=, <=
//              for all combinations of C-style strings and String
//
//  concatenation operators and functions
//          String & operator +=(const String & x)
//              catenates x to the right of the String
//          friend String operator +(const String & x, const String & y)
//              catenates x to y and returns result
//
//
//  search/facilitator functions
//
//          String substr(int pos, int len)
//             returns String of len chars starting at pos
//             truncates if len too big, returns empty String
//             if pos too big
//
//          String Downcase() const
//          String Upcase() const
//               return lower/upper case versions of String, respectively
//
//          int Contains(const String & s) const
//          int Contains(char * s) const
//               returns 1 if String contains s, else returns 0
//
// conversion functions
//
//          char * chars()
//             similar to (char *) cast, but member function
//
//          int atoi(String &)
//             like <stdlib.h> atoi, but has String parameter
//             ensures that calling <stdlib.h> version is done
//             correctly
//
//          int atof(String &)
//             like <stdlib.h> atof, but has String parameter
//             ensures that calling <stdlib.h> version is done
//             correctly


class String{
    
  public:
    String(const char * s = 0)
    // precondition: s is '\0'-terminated C-style string
    // postcondition: object constructed to have same value as s
    {
        if (s != 0){                            // C-string not NULL 
            myLength = strlen(s) + 1;          // make room for '\0'  
            myCstring = new char[myLength];
            assert(myCstring != 0);
            strcpy(myCstring,s);
        }
        else{
            myLength = 1;
            myCstring = new char[1];
            assert(myCstring != 0);
            myCstring[0] = '\0';
        }
        
    }   


    String(const String & x)
    {
        if (this != &x){
            myLength = x.length() + 1;
            myCstring = new char[myLength];
            assert(myCstring != 0);
            strcpy(myCstring,x.myCstring);
        }
    }

    String(char ch)
    {
        myLength = 2;
        myCstring = new char[myLength];
        assert(myCstring != 0);
        myCstring[0] = ch;
        myCstring[1] = '\0';
    }



    String& operator = (const char * s)
    // precondition: s is '\0'-terminated C-style string
    // postcondition: assign object the value of C-style string s
    {

        int len = 0;          // length of newly constructed String
        if (s != 0){
            len = strlen(s);  // assume '\0' terminated
        }
        
        // free old String if necessary
        
        if (myLength < len + 1){
            delete [] myCstring;
            myLength = len + 1;
            myCstring = new char[myLength];
            assert(myCstring != 0);
        }

        strcpy(myCstring,s);
        return *this;
    }

    String& operator =(const String & x)
    {
        if (this != &x){
            if (myLength < x.length() + 1){       // need more chars
                delete[] myCstring;                  // delete old space
                myLength = x.length() + 1;
                myCstring = new char[myLength];     // get new space
                assert(myCstring != 0);
            }
            strcpy(myCstring,x.myCstring);
        }
        return *this;   
    }

    ~String()
    {
        delete[] myCstring;
        myCstring = 0;
        myLength = 0;
    }
    
    operator const char *() const {return myCstring;}

    char * chars() {return myCstring;}

    int length() const {return strlen(myCstring);}

    char& operator[](int i)
    {
        if (i < 0 || strlen(myCstring) <= i){
            cerr << "index out of range: " << i << " string: " << myCstring
                << endl;
            assert(0 <= i && i < strlen(myCstring));
        }
        return myCstring[i];
    }
    const char& operator[](int i) const
    {
        if (i < 0 || strlen(myCstring) <= i){
            cerr << "index out of range: " << i << " string: " << myCstring
                << endl;
            assert(0 <= i && i < strlen(myCstring));
        }
        return myCstring[i];
    }
        

    friend ostream& operator <<(ostream& os, const String& str)
    {
        return os << str.myCstring;
    }

    friend istream& operator >>(istream& is, String& str)
    {
        char buf[String::maxLength];
        is >> buf;
        str = buf;
        return is;
    }


    friend istream & getline(istream & is, String & s, char sentinel = '\n')
    {
        char buf[String::maxLength+1];
        is.getline(buf,maxLength, sentinel);
        if (s.myLength < strlen(buf) + 1) {
            delete [] s.myCstring;
            s.myLength = strlen(buf) + 1;
            s.myCstring = new char[s.myLength];
            assert(s.myCstring != 0);
        }
        
        strcpy(s.myCstring,buf);
        
        return is;
    }

    istream & GetLine(istream & is, char sentinel = '\n')
    {
        char buf[String::maxLength+1];
        is.getline(buf,maxLength, sentinel);
        if (myLength < strlen(buf) + 1)
        {
            delete [] myCstring;
            myLength = strlen(buf) + 1;
            myCstring = new char[myLength];
            assert(myCstring != 0);
        }

        strcpy(myCstring,buf);

        return is;
    }


    String& operator +=(const String & x)
    // postcondition: concatenates x onto object
    {
        int newLength = length() + x.length();       // object + added String

           // check to see if local buffer not big enough
        if (newLength >= myLength){
            char * newBuffer = new char[newLength + 1];
            assert(newBuffer != 0);

            strcpy(newBuffer,myCstring);               // copy into new buffer
            delete [] myCstring;                       // delete old String
            myLength = newLength + 1;               // update information
            myCstring = newBuffer;
        }
           // now catenate x to end of myCstring

        int lastLocation = length();       // location of '\0'
        strcpy(myCstring+lastLocation,x);

        return *this;
    }

    friend String operator +(const String & x, const String & y)
    {
        String result(x);       // copies x to result
        result += y;            // catenate y
        return result;          // returns a copy of result
    }


    String Downcase() const
    {
        String result = myCstring;
        int len = length();
        int k;
        for(k=0; k < len; k++)
        {
            result[k] = tolower(result[k]);
        }
        return result;  
    }

    String Upcase() const
    {
        String result = myCstring;
        int len = length();
        int k;
        for (k=0; k < len; k++) {
            result[k] = toupper(result[k]);
        }
        return result;  
    }
    
    String substr(int pos, int len)
    // precondition: 0 <= pos < s.length() && pos + len <= s.length()
    // postcondition: returns substring starting at char pos of len chars
    {
        String result = "";
        int mylen = length();
        int lastIndex = pos + len - 1;     // last char's index (to copy)

        if (pos >= mylen) return result;   // empty String

        if (lastIndex >= mylen) {          // off end of String?
            lastIndex = mylen-1;
        }

        int k;
        for (k=pos; k <= lastIndex; k++) {
            result += myCstring[k];
        }

        return result;
    }

    // Phil's condition if no length is give, returns the rest of it.
    String substr(int pos)
    // precondition: 0 <= pos < s.length()
    // postcondition: returns substring starting at char pos
    {
        String result = "";
        int mylen = length();

        if (pos >= mylen) return result;   // empty String

        int lastIndex = mylen-1;

        int k;
        for (k=pos; k <= lastIndex; k++) {
            result += myCstring[k];
        }

        return result;
    }

    
    int Contains (const String & s) const
    {
        return Contains(s.myCstring);
    }

    int Contains(char * s) const
    {
        int len = strlen(s);
        int lastIndex = length() - len;
        int k;
        for (k=0; k <= lastIndex; k++) {
            if (strncmp(myCstring + k,s,len) == 0) return 1;
        }
        return 0;
    }


// relational operators for String/String comparisons

    friend int operator ==(const String & x, const String & y)
        {return strcmp(x.myCstring, y.myCstring)==0;}

    friend int operator !=(const String & x, const String & y)
        {return ! (x == y);}

    friend int operator < (const String & x, const String & y)
         {return strcmp(x.myCstring, y.myCstring) < 0;}

    friend int operator > (const String & x, const String & y)
         {return strcmp(x.myCstring, y.myCstring) > 0;}

    friend int operator >= (const String & x, const String & y)
         {return x > y || x == y;}

    friend int operator <= (const String & x, const String & y)
         {return x < y || x == y;}

// relational operators for String/char * comparisons

    friend int operator ==(const String & x, const char * s)
        {return strcmp(x.myCstring,s)==0;}

    friend int operator !=(const String & x, const char * s)
        {return ! (x == s);}

    friend int operator < (const String & x, const char * s)
          {return strcmp(x.myCstring, s) < 0;}

    friend int operator > (const String & x, const char *s)
          {return strcmp(x.myCstring, s) > 0;}

    friend int operator >= (const String & x, const char *s)
         {return x > s || x == s;}

    friend int operator <= (const String & x, const char *s)
         {return x < s || x == s;}


// relational operators for char */String comparisons

    friend int operator ==(const char * s, const String & x)
        {return strcmp(s, x.myCstring)==0;}

    friend int operator !=(const char * s, const String & x)
        {return ! (s == x);}

    friend int operator < (const char * s, const String & x)
          {return strcmp(s,x.myCstring) < 0;}

    friend int operator > (const char *s, const String & x)
          {return strcmp(s,x.myCstring) > 0;}

    friend int operator >= (const char *s, const String & x)
         {return s > x || s == x;}

    friend int operator <= (const char *s, const String & x)
         {return s < x || s == x;}

  private:
    char * myCstring;         // private C-style string
    enum {maxLength = 1024};  // maximum length of string read from input
    int myLength;


};

inline int atoi(String & s)
// precondition: s is a sequence of digits
// postcondition: returns integer equivalent of s
// exception: if s is not valid integer, 0 is returned
{
    return atoi(s.chars());
}

inline double atof(String & s)
// precondition: s is a sequence of digits
// postcondition: returns double equivalent of s
// exception: if s is not valid double, 0 is returned
{
    return atof(s.chars());
}

#endif /* _MYSTRING_H not defined */
